import gwyfile
import gsffile
import numpy as np

class Reader:
    def __init__(self, fullfilepath=None):
        self.filename = fullfilepath

class GwyReader(Reader):

    def __init__(self, fullfilepath=None, channelname=None):
        super().__init__(fullfilepath)
        self.channelname = channelname

    def read(self):
        # Returns a dictionary of all the channels
        gwyobj = gwyfile.load(self.filename)
        allchannels = gwyfile.util.get_datafields(gwyobj)

        if self.channelname == None:
            return allchannels
        else:
            # Read channels from gwyfile and return only a specific one
            channel = allchannels[self.channelname]
            return channel

class GsfReader(Reader):
    def __init__(self, fullfilepath=None):
        super().__init__(fullfilepath)

    def read(self):
        data, metadata = gsffile.read_gsf(self.filename)
        channel = gwyfile.objects.GwyDataField(data,
                 xreal=metadata["XReal"], yreal=metadata["YReal"], xoff=metadata["XOffset"], yoff=metadata["YOffset"],
                 si_unit_xy=None, si_unit_z=None,
                 typecodes=None)
        return channel

class NeaInfoReader(Reader):
    def __init__(self, fullfilepath=None):
        super().__init__(fullfilepath)

    def read(self):
        # reader tested for neascan version 2.1.10719.0
        fid = open(self.filename, errors='replace')
        infodict = {}

        linestring = ''
        Nlines = 0

        while 'Version:' not in linestring:
            Nlines += 1
            linestring = fid.readline()
            if Nlines > 1:
                ct = linestring.split('\t')
                fieldname = ct[0][2:-1]
                fieldname = fieldname.replace(' ', '')

                if 'Scanner Center Position' in linestring:
                    fieldname = fieldname[:-5]
                    infodict[fieldname] = [float(ct[2]), float(ct[3])]

                elif 'Scan Area' in linestring:
                    fieldname = fieldname[:-7]
                    infodict[fieldname] = [float(ct[2]), float(ct[3]), float(ct[4])]

                elif 'Pixel Area' in linestring:
                    fieldname = fieldname[:-7]
                    infodict[fieldname] = [int(ct[2]), int(ct[3]), int(ct[4])]

                elif 'Interferometer Center/Distance' in linestring:
                    fieldname = fieldname.replace('/', '')
                    infodict[fieldname] = [float(ct[2]), float(ct[3])]

                elif 'Regulator' in linestring:
                    fieldname = fieldname[:-7]
                    infodict[fieldname] = [float(ct[2]), float(ct[3]), float(ct[4])]

                elif 'Q-Factor' in linestring:
                    fieldname = fieldname.replace('-', '')
                    infodict[fieldname] = float(ct[2])

                else:
                    fieldname = ct[0][2:-1]
                    fieldname = fieldname.replace(' ', '')
                    val = ct[2]
                    val = val.replace(',', '')
                    try:
                        infodict[fieldname] = float(val)
                    except:
                        infodict[fieldname] = val.strip()
        fid.close()
        return infodict

class NeaSpectrumReader(Reader):
    def __init__(self, fullfilepath=None):
        super().__init__(fullfilepath)

    def read(self):
        # reader tested for neascan version 2.1.10719.0
        fid = open(self.filename,errors='replace')
        data = {}
        params = {}

        linestring = fid.readline()
        Nlines = 1

        while 'Row' not in linestring:
            Nlines += 1
            linestring = fid.readline()
            if Nlines > 1:
                ct = linestring.split('\t')
                fieldname = ct[0][2:-1]
                fieldname = fieldname.replace(' ', '')

                if 'Scanner Center Position' in linestring:
                    fieldname = fieldname[:-5]
                    params[fieldname] = [float(ct[2]), float(ct[3])]

                elif 'Scan Area' in linestring:
                    fieldname = fieldname[:-7]
                    params[fieldname] = [float(ct[2]), float(ct[3]), float(ct[4])]

                elif 'Pixel Area' in linestring:
                    fieldname = fieldname[:-7]
                    params[fieldname] = [int(ct[2]), int(ct[3]), int(ct[4])]

                elif 'Interferometer Center/Distance' in linestring:
                    fieldname = fieldname.replace('/', '')
                    params[fieldname] = [float(ct[2]), float(ct[3])]

                elif 'Regulator' in linestring:
                    fieldname = fieldname[:-7]
                    params[fieldname] = [float(ct[2]), float(ct[3]), float(ct[4])]

                elif 'Q-Factor' in linestring:
                    fieldname = fieldname.replace('-', '')
                    params[fieldname] = float(ct[2])

                else:
                    fieldname = ct[0][2:-1]
                    fieldname = fieldname.replace(' ', '')
                    val = ct[2]
                    val = val.replace(',','')
                    try:
                        params[fieldname] = float(val)
                    except:
                        params[fieldname] = val.strip()

        channels = linestring.split('\t')
        fid.close()

        if "PTE+" in params['Scan']:
            C_data = np.genfromtxt(self.filename, skip_header=Nlines, encoding='utf-8')
        else:
            C_data = np.genfromtxt(self.filename, skip_header=Nlines)

        for i in range(len(channels)-1):
            data[channels[i]] = C_data[:,i]

        return data, params
    
class NeaInterferogramReader(Reader):
    def __init__(self, fullfilepath=None):
        super().__init__(fullfilepath)

    def read(self):
        # reader tested for neascan version 2.1.10719.0
        fid = open(self.filename,errors='replace')
        data = {}
        params = {}

        linestring = fid.readline()
        Nlines = 1

        while 'Row' not in linestring:
            Nlines += 1
            linestring = fid.readline()
            if Nlines > 1:
                ct = linestring.split('\t')
                fieldname = ct[0][2:-1]
                fieldname = fieldname.replace(' ', '')

                if 'Scanner Center Position' in linestring:
                    fieldname = fieldname[:-5]
                    params[fieldname] = [float(ct[2]), float(ct[3])]

                elif 'Scan Area' in linestring:
                    fieldname = fieldname[:-7]
                    params[fieldname] = [float(ct[2]), float(ct[3]), float(ct[4])]

                elif 'Pixel Area' in linestring:
                    fieldname = fieldname[:-7]
                    params[fieldname] = [int(ct[2]), int(ct[3]), int(ct[4])]

                elif 'Averaging' in linestring:
                    # fieldname = fieldname[:-7]
                    params[fieldname] = int(ct[2])

                elif 'Interferometer Center/Distance' in linestring:
                    fieldname = fieldname.replace('/', '')
                    params[fieldname] = [float(ct[2]), float(ct[3])]

                elif 'Regulator' in linestring:
                    fieldname = fieldname[:-7]
                    params[fieldname] = [float(ct[2]), float(ct[3]), float(ct[4])]

                elif 'Q-Factor' in linestring:
                    fieldname = fieldname.replace('-', '')
                    params[fieldname] = float(ct[2])

                else:
                    fieldname = ct[0][2:-1]
                    fieldname = fieldname.replace(' ', '')
                    val = ct[2]
                    val = val.replace(',','')
                    try:
                        params[fieldname] = float(val)
                    except:
                        params[fieldname] = val.strip()

        channels = linestring.split('\t')
        fid.close()

        C_data = np.genfromtxt(self.filename, skip_header=Nlines)

        for i in range(len(channels)-1):
            data[channels[i]] = C_data[:,i]
        
        return data, params


import unittest

import numpy as np

from pySNOM.images import LineLevel, DataTypes


class TestLineLevel(unittest.TestCase):

    def test_median(self):
        d = np.arange(12).reshape(3, -1)[:, [0, 1, 3]]
        l = LineLevel(method="median", datatype=DataTypes.Phase)
        out = l.transform(d)
        np.testing.assert_almost_equal(out, [[-1.,  0.,  2.],
                                             [-1.,  0.,  2.],
                                             [-1.,  0.,  2.]])
        l = LineLevel(method="median", datatype=DataTypes.Amplitude)
        out = l.transform(d)
        np.testing.assert_almost_equal(out, [[0.       , 1.       , 3.       ],
                                             [0.8      , 1.       , 1.4      ],
                                             [0.8888889, 1.       , 1.2222222]])

    def test_mean(self):
        d = np.arange(12).reshape(3, -1)[:, [0, 1, 3]]
        l = LineLevel(method="mean", datatype=DataTypes.Phase)
        out = l.transform(d)
        np.testing.assert_almost_equal(out, [[-1.3333333, -0.3333333,  1.6666667],
                                             [-1.3333333, -0.3333333,  1.6666667],
                                             [-1.3333333, -0.3333333,  1.6666667]])
        l = LineLevel(method="mean", datatype=DataTypes.Amplitude)
        out = l.transform(d)
        np.testing.assert_almost_equal(out, [[0.       , 0.75     , 2.25     ],
                                             [0.75     , 0.9375   , 1.3125   ],
                                             [0.8571429, 0.9642857, 1.1785714]])

    def test_difference(self):
        d = np.arange(12).reshape(3, -1)[:, [0, 1, 3]]
        l = LineLevel(method="difference", datatype=DataTypes.Phase)
        out = l.transform(d)
        np.testing.assert_almost_equal(out, [[-4., -3., -1.],
                                             [ 0.,  1.,  3.]])
        l = LineLevel(method="difference", datatype=DataTypes.Amplitude)
        out = l.transform(d)
        np.testing.assert_almost_equal(out, [[0.       , 0.2      , 0.6      ],
                                             [2.2222222, 2.7777778, 3.8888889]])


if __name__ == '__main__':
    unittest.main()


import numpy as np
from enum import Enum
from pySNOM.images import type_from_channelname
from pySNOM.defaults import Defaults

MeasurementModes = Enum('MeasurementModes', ['None','nanoFTIR', 'PsHet', 'PTE'])
DataTypes = Enum('DataTypes', ['Amplitude', 'Phase', 'Complex', 'Topography'])
ChannelTypes = Enum('ChannelTypes', ['None','Optical','Mechanical'])
ScanTypes = Enum('ScanTypes',['Point','LineScan','HyperScan'])

class NeaSpectrum:
    """Storing full spectral measurement data and corresponding measurement parameters"""
    def __init__(self, data: dict, parameters: dict, scantype="Point", filename=None, mode="nanoFTIR"):
        self.filename = filename # Full path with name
        self._parameters = parameters
        self.data = data
        
        # set measurement mode
        try:
            self._mode = MeasurementModes[mode]
        except ValueError:
            self._mode = MeasurementModes["nanoFTIR"]
            raise ValueError(mode + 'is not a measurement mode!')
        
        try:
            self._scantype = ScanTypes[scantype]
        except ValueError:
            self._scantype = ScanTypes["Point"]
            raise ValueError(scantype + 'is not a measurement mode!')
        
        if parameters:
            try:
                if parameters['PixelArea'][1] == 1 and parameters['PixelArea'][0] == 1:
                    self._scantype = ScanTypes["Point"]
                elif parameters['PixelArea'][1] == 1 or parameters['PixelArea'][0] == 1:
                    self._scantype = ScanTypes["LineScan"]
                else:
                    self._scantype = ScanTypes["HyperScan"]
                self._mode = MeasurementModes[Defaults().spectral_mode_defs[parameters["Scan"]]]
            except:
                raise ValueError('Parameters dictionary is not valid!')
        else:
            # set measurement mode
            try:
                self._mode = MeasurementModes[mode]
            except ValueError:
                self._mode = MeasurementModes["nanoFTIR"]
                raise ValueError(mode + 'is not a measurement mode!')
            # set scan type
            try:
                self._scantype = ScanTypes[scantype]
            except ValueError:
                self._scantype = ScanTypes["Point"]
                raise ValueError(scantype + 'is not a measurement mode!')

    @property
    def data(self):
        """Property - data (dict with measurement channels)"""
        return self._data
    
    @data.setter
    def data(self, value):
        """Data setter to reshape properly"""
        self._data = Tools.reshape_spectrum_data(value, self._parameters)
    
    @property
    def mode(self):
        """Property - mode (MeasurementMode Enum name)"""
        return self._mode.name
        
    @property
    def scantype(self):
        """Property - scantype (ScanType Enum name)"""
        return self._scantype.name
    
    @property
    def parameters(self):
        """Property - scantype (ScanType Enum name)"""
        return self._parameters

class SingleChannelSpectrum(NeaSpectrum):
    def __init__(self, data, wndata, filename=None, parameters=None, mode="nanoFTIR", channelname='O2A'):
        super().__init__(data, filename, parameters, mode)
        self._wndata = wndata
        self.channel = channelname
        self.order = 0
        self.datatype = None

    @property
    def data(self):
        """Property - data (dict with measurement channels)"""
        return self._data
    
    @property
    def wndata(self):
        """Property - wavenumber data"""
        return self._wndata

    @property
    def channel(self):
        """Property - channel (string)"""
        return self._channel
    
    @channel.setter
    def channel(self,value):

        self._channel = value
        self.channeltype, self.order, self.datatype = type_from_channelname(value)

# TRANSFORMATIONS ------------------------------------------------------------------------------------------------------------------
class Transformation:

    def transform(self, data):
        raise NotImplementedError()
    
class LinearNormalize(Transformation):

    def __init__(self, wavenumber1=0.0, wavenumber2=1000.0, datatype=DataTypes.Phase):
        self.wn1 = wavenumber1
        self.wn2 = wavenumber2
        self.datatype = datatype

    def transform(self, spectrum, wnaxis):
        wn1idx = np.argmin(abs(wnaxis - self.wn1))
        wn2idx = np.argmin(abs(wnaxis - self.wn2))
        m = (spectrum[wn2idx] - spectrum[wn1idx])/(wnaxis[wn2idx] - wnaxis[wn1idx])
        C = spectrum[wn1idx] - m*wnaxis[wn1idx]

        if self.datatype == DataTypes.Amplitude:
            return spectrum / (m*wnaxis + C)
        else:
            return spectrum - (m*wnaxis + C)

class RotatePhase(Transformation):

    def __init__(self, degree=0.0, wn_ref=1000.0):
        self.wn_ref = wn_ref
        self.degree = degree

    def transform(self, spectrum, wnaxis):
        if not np.iscomplex(spectrum).any():
            spectrum = np.exp(spectrum*complex(1j))

        angles = wnaxis * np.deg2rad(self.degree) / self.wn_ref

        return np.angle(spectrum*np.exp(angles*complex(1j)))
    
class NormalizeSpectrum(Transformation):

    def __init__(self, datatype=DataTypes.Phase, dounwrap=False):
        self.datatype = datatype
        self.dounwrap = dounwrap

    def transform(self, spectrum, refspectrum):
        if self.datatype == DataTypes.Phase or self.datatype == DataTypes.Topography:
            if self.dounwrap:
                return np.unwrap(spectrum - refspectrum)
            else:
                return spectrum - refspectrum
        else:
            return spectrum / refspectrum

# TOOLS ------------------------------------------------------------------------------------------------------------------
class Tools:
    def reshape_spectrum_data(data, params):
        for channel in list(data.keys()):
            # Point spectrum
            if params['PixelArea'][1] == 1 and params['PixelArea'][0] == 1:
                # NeaSpec tends to do zero-filling: 4 when processing spectra
                if params['Scan'] == "Fourier Scan":
                    data[channel] = np.reshape(data[channel], (params['PixelArea'][2]*2))
                else:
                    data[data[channel]] = np.reshape(data[channel], (params['PixelArea'][2]))
            # LineScan
            elif params['PixelArea'][1] == 1 and params['PixelArea'][0] != 1:
                # NeaSpec tends to do zero-filling: 4 when processing spectra
                if params['Scan'] == "Fourier Scan":
                    data[data[channel]] = np.reshape(data[channel], (params['PixelArea'][0], params['PixelArea'][2]*2))
                else:
                    data[data[channel]] = np.reshape(data[channel], (params['PixelArea'][0], params['PixelArea'][2]))
            # HyperScan
            else:
                # NeaSpec tends to do zero-filling: 4 when processing spectra
                if params['Scan'] == "Fourier Scan":
                    data[data[channel]] = np.reshape(data[channel], (params['PixelArea'][0], params['PixelArea'][1], params['PixelArea'][2]*2))
                else:
                    data[data[channel]] = np.reshape(data[channel], (params['PixelArea'][0], params['PixelArea'][1], params['PixelArea'][2]))

        return data


import unittest
import os

import numpy as np

import pySNOM
from pySNOM import readers
from pySNOM.interferograms import NeaInterferogram, ProcessSingleChannel, ProcessMultiChannels, ProcessAllPoints

class test_Neaspectrum(unittest.TestCase):

    def __init__(self, methodName: str = "runTest") -> None:
        super().__init__(methodName)
        fdata = 'datasets/testifg_singlepoint.txt'
        data_reader = readers.NeaInterferogramReader(os.path.join(pySNOM.__path__[0], fdata))
        data, measparams = data_reader.read()

        self.ifg = NeaInterferogram(data,measparams,filename=fdata)

    def test_pointinterferogram_object(self):

        np.testing.assert_almost_equal(self.ifg.data['O2A'][100], 9.564073)
        np.testing.assert_string_equal(self.ifg.parameters["Scan"], "Fourier Scan")
        np.testing.assert_string_equal(self.ifg.scantype, "Point")
        np.testing.assert_equal(np.shape(self.ifg.data["O2A"])[0],46080)

    def test_singlechannel_process(self):

        a2, p2, wn2 = ProcessSingleChannel(order=2,simpleoutput=True).transform(self.ifg)

        np.testing.assert_almost_equal(a2[500],65.082045184048)
        np.testing.assert_almost_equal(p2[500],-2.2006946866734887)
        np.testing.assert_almost_equal(wn2[500],1273.6715150986495)

    def test_multichannel_process(self):
        s_multichannel = ProcessMultiChannels(apod=True).transform(self.ifg)

        np.testing.assert_almost_equal(s_multichannel.data["O3A"][500],24.262286272717716)
        np.testing.assert_almost_equal(s_multichannel.data["O3P"][500],0.3032580222572598)
        np.testing.assert_almost_equal(s_multichannel.data["Wavenumber"][500],1273.6715150986495)

    def test_fullauto_process(self):
        s_allpoints = ProcessAllPoints().transform(self.ifg)

        np.testing.assert_almost_equal(s_allpoints.data["O3A"][500],24.262286272717716)
        np.testing.assert_almost_equal(s_allpoints.data["O3P"][500],0.3032580222572598)
        np.testing.assert_almost_equal(s_allpoints.data["Wavenumber"][500],1273.6715150986495)

if __name__ == '__main__':
    unittest.main()


import numpy as np
import copy
from enum import Enum
from scipy import signal
from scipy.fft import fft, fftshift
from scipy.interpolate import CubicSpline, interp1d
from pySNOM.spectra import NeaSpectrum, SingleChannelSpectrum

MeasurementModes = Enum('MeasurementModes', ['None','nanoFTIR'])
DataTypes = Enum('DataTypes', ['Amplitude', 'Phase', 'Topography'])
ChannelTypes = Enum('ChannelTypes', ['None','Optical','Mechanical'])
ScanTypes = Enum('ScanTypes',['Point','LineScan','HyperScan'])

# INTERFEROGRAMS ------------------------------------------------------------------------------------------------------------------
class NeaInterferogram(NeaSpectrum):
    def __init__(self, data, parameters, scantype="Point", filename=None, mode="nanoFTIR"):
        super().__init__(data, parameters, scantype=scantype, filename=filename, mode=mode)
        self.data = data

    @property
    def data(self):
        """Property - data (dict with measurement channels)"""
        return self._data

    @data.setter
    def data(self, value):
        """Data setter to reshape properly"""
        self._data = Tools.reshape_ifg_data(value, self._parameters)

# TRANSFORMATIONS ------------------------------------------------------------------------------------------------------------------
class Transformation:

    def transform(self, data):
        raise NotImplementedError()

class ProcessInterferogram(Transformation):

    def __init__(self, apod = True, windowtype = "blackmanharris", nzeros = 4, wlpidx = None):
        self.apod = apod
        self.nzeros = nzeros
        self.wlpidx = wlpidx
        self.windowtype = windowtype

    def transform(self, ifg, maxis):
        # Find the location index of the WLP
        if self.wlpidx is None:
            self.wlpidx = np.argmax(np.abs(ifg))
        # Create apodization window
        if self.apod:
            w = Tools.asymmetric_window(npoints = len(ifg), centerindex = self.wlpidx, windowtype = self.windowtype)
        else:
            w = np.ones(np.shape(ifg))
        ifg = ifg - np.mean(ifg)
        # Calculate FFT
        complex_spectrum = fftshift(fft(ifg*w, self.nzeros*len(ifg)))
        # Calculate frequency axis
        stepsizes = np.mean(np.diff(maxis*1e6))
        Fs = 1/np.mean(stepsizes)
        faxis = (Fs/2)*np.linspace(-1,1,len(complex_spectrum))*10000/2

        return complex_spectrum[int(len(faxis)/2)-1:-1], faxis[int(len(faxis)/2)-1:-1]

class InterpolateInterferogram(Transformation):
        """Reinterpolates the raw interferograms to a uniform grid."""
        def __init__(self, method = "spline"):
            self.method = method

        def transform(self, ifg, maxis):
            # if np.iscomplex(ifg).any():
            newifg = np.zeros(np.shape(ifg))*complex(1j)
            # else:
            # newifg = np.zeros(np.shape(ifg))
            
            newmaxis = np.zeros(np.shape(maxis))

            match self.method:
                case "spline":
                    interp_object = CubicSpline
                case "linear":
                    interp_object = interp1d

            # in case of processing multiple interferograms in a 2d array we have to take the median
            # sometimes the point locations has large jumps at the beginning (neaspec machine artifact)
            startM = np.min(np.nanmedian(maxis,axis=0))
            stopM = np.max(np.nanmedian(maxis,axis=0))

            if ifg.ndim == 1:
                newcoords = np.linspace(startM,stopM,num = len(maxis))
                if np.iscomplex(ifg).any():
                    interpR = interp_object(maxis,np.real(ifg))
                    interpI = interp_object(maxis,np.imag(ifg))
                    newifgR = interpR(newcoords)
                    newifgI = interpI(newcoords)
                    newifg = newifgR + newifgI * complex(1j)
                else:
                    interpifg = interp_object(maxis,ifg)
                    newifg = interpifg(newcoords)
                return newifg, newcoords
            elif ifg.ndim == 2:
                newcoords = np.linspace(startM,stopM,num=np.shape(maxis)[1])
                for i in range(np.shape(ifg)[0]):
                    if np.iscomplex(ifg[i][:]).any():
                        interpR = interp_object(maxis[i][:],np.real(ifg[i][:]))
                        interpI = interp_object(maxis[i][:],np.imag(ifg[i][:]))
                        newifgR = interpR(newcoords)
                        newifgI = interpI(newcoords)
                        newifg[i][:] = newifgR + newifgI * complex(1j)
                        newmaxis[i][:] = newcoords
                    else:
                        interpifg = interp_object(maxis[i][:], ifg[i][:])
                        newifg[i][:] = interpifg(newcoords)
                        newmaxis[i][:] = newcoords
                return newifg, newmaxis
            else:
                return ifg, maxis

class ProcessSingleChannel(Transformation):
    def __init__(self, order, method = "complex", apod = True, windowtype = "blackmanharris", nzeros = 4, interpmethod = "spline", simpleoutput = False):
        self.order = order
        self.method = method
        self.windowtype = windowtype
        self.apod = apod
        self.nzeros = nzeros
        self.interpmethod = interpmethod
        self.simpleoutput = simpleoutput

    def transform(self, neaifg):# Load amplitude and phase of the given channel
            
        # Calculate the interferogram to process based on the given method

        channelA = f"O{self.order}A"
        channelP = f"O{self.order}P"

        ifgA = np.reshape(neaifg.data[channelA],(neaifg.parameters['Averaging'],neaifg.parameters['PixelArea'][2]))
        ifgP = np.reshape(neaifg.data[channelP],(neaifg.parameters['Averaging'],neaifg.parameters['PixelArea'][2]))
        Maxis = np.reshape(neaifg.data["M"],(neaifg.parameters['Averaging'],neaifg.parameters['PixelArea'][2]))

        match self.method:
            case "abs":
                IFG = np.abs(ifgA*np.exp(ifgP*complex(1j)))
            case "real":
                IFG = np.real(ifgA*np.exp(ifgP*complex(1j)))
            case "imag":
                IFG = np.imag(ifgA*np.exp(ifgP*complex(1j)))
            case "complex":
                IFG = ifgA*np.exp(ifgP*complex(1j))
            case "simple":
                IFG = ifgA

        #  Interpolate
        IFG, Maxis = InterpolateInterferogram(method = self.interpmethod).transform(IFG, Maxis)
        
        # PROCESS IFGs
        # Check if it is multiple interferograms or just a single one
        if len(np.shape(IFG)) == 1:
            complex_spectrum, f = ProcessInterferogram(apod = self.apod, windowtype = self.windowtype, nzeros = self.nzeros).transform(IFG,Maxis)
            amp = np.abs(complex_spectrum)
            phi = np.angle(complex_spectrum)
        else:
            # Allocate variables
            spectraAll = complex(1j)*np.zeros((np.shape(IFG)[0],int(self.nzeros*np.shape(IFG)[1]/2)))
            fAll = np.zeros(np.shape(spectraAll))
            # Go trough all
            for i in range(np.shape(IFG)[0]):
                spectraAll[i,:], fAll[i,:] = ProcessInterferogram(apod = self.apod, windowtype = self.windowtype, nzeros = self.nzeros).transform(IFG[i,:], Maxis[i,:])
            # Average the complex spectra
            complex_spectrum = np.mean(spectraAll, axis = 0)
            # Extract amplitude and phase from the averaged complex spectrum
            amp = np.abs(complex_spectrum)
            phi = np.angle(complex_spectrum)
            f = np.mean(fAll, axis=0)

        if self.simpleoutput:
            return amp, phi, f
        else:
            spectrum_data = {}
            spectrum_parameters = copy.deepcopy(neaifg.parameters)
            spectrum_parameters["ScanArea"] = [neaifg.parameters["ScanArea"][0],neaifg.parameters["ScanArea"][1],len(amp)]
            spectrum_data[channelA] = amp
            spectrum_data[channelP] = phi
            spectrum_data["Wavenumber"] = f
            spectrum = NeaSpectrum(spectrum_data, spectrum_parameters)

            return spectrum

class ProcessMultiChannels(Transformation):

    def __init__(self, method = "complex", apod = True, windowtype = "blackmanharris", nzeros = 4, interpmethod = "spline", simpleoutput = False):
        self.method = method
        self.apod = apod
        self.windowtype = windowtype
        self.nzeros = nzeros
        self.interpmethod = interpmethod
        self.simpleoutput = simpleoutput

    def transform(self, neaifg):
        spectrum_data = {}
        spectrum_parameters = copy.deepcopy(neaifg.parameters)

        for order in range(6):

            channelA = f"O{order}A"
            channelP = f"O{order}P"

            amp, phi, f = ProcessSingleChannel(order, method = self.method, apod = self.apod, windowtype = self.windowtype, nzeros = self.nzeros, interpmethod = self.interpmethod, simpleoutput=True).transform(neaifg)
            
            spectrum_data[channelA] = amp
            spectrum_data[channelP] = phi
            spectrum_data["Wavenumber"] = f

        if self.simpleoutput:
            spectrum_data
        else:
            spectrum_parameters["ScanArea"] = [neaifg.parameters["ScanArea"][0],neaifg.parameters["ScanArea"][1],len(amp)]
            spectrum = NeaSpectrum(spectrum_data, spectrum_parameters)
            return spectrum

class ProcessAllPoints(Transformation):

    def __init__(self, method = "complex", apod = True, windowtype = "blackmanharris", nzeros = 4, interpmethod = "spline"):
        self.method = method
        self.apod = apod
        self.windowtype = windowtype
        self.nzeros = nzeros
        self.interpmethod = interpmethod
        
    def transform(self,neaifg):
        if neaifg.parameters['PixelArea'][0] == 1 and neaifg.parameters['PixelArea'][1] == 1:
            spectra = ProcessMultiChannels(method = self.method, windowtype = self.windowtype, nzeros = self.nzeros, 
                                           apod = self.apod, interpmethod = self.interpmethod, simpleoutput = False).transform(neaifg)
        else:
            pixel_area = [neaifg.parameters["PixelArea"][0],neaifg.parameters["PixelArea"][1],int(self.nzeros*neaifg.parameters['PixelArea'][2]/2)]

            ampFullData = np.zeros((pixel_area[0], pixel_area[1], pixel_area[2]))
            phiFullData = np.zeros(np.shape(ampFullData))
            fFullData = np.zeros(np.shape(ampFullData))

            pointifg_data = dict()
            pointifg_params = dict()
            pointifg_params["PixelArea"] = [1,1,neaifg.parameters["PixelArea"][2]]

            spectra = NeaSpectrum({},{}, scantype=neaifg.scantype)
            
            for order in range(6):
                channelA = f"O{order}A"
                channelP = f"O{order}P"
                for i in range(pixel_area[0]):
                    for k in range(pixel_area[1]):
                            pointifg_data[channelA] = neaifg.data[channelA][i,k,:]
                            pointifg_data[channelP] = neaifg.data[channelP][i,k,:]
                            pointifg_data["M"] = neaifg.data["M"][i,k,:]

                            ampFullData[i,k,:], phiFullData[i,k,:], fFullData[i,k,:] = ProcessSingleChannel(order, method = self.method, apod = self.apod, 
                                                                                                            windowtype = self.windowtype, nzeros = self.nzeros, 
                                                                                                            interpmethod = self.interpmethod, simpleoutput=True).transform(neaifg)                    
                
                spectra.data[channelA] = ampFullData
                spectra.data[channelP] = phiFullData
                spectra.data["Wavenumber"] = fFullData
            
            spectra._parameters["PixelArea"] = pixel_area

        return spectra

# TOOLS ------------------------------------------------------------------------------------------------------------------
class Tools:
    def __init__(self):
        pass
    
    def reshape_ifg_data(data, params):
        if params['PixelArea'][1] != 1 and params['PixelArea'][0] == 1:
            for channel in list(data.keys()):
                data[channel] = np.reshape(data[channel], (int(params['PixelArea'][0]), int(params['PixelArea'][1]), int(params['PixelArea'][2]*params['Averaging'])))
            return data
        else:
            return data

    def reshape_linescan_interferogram(data, parameters):
        return np.reshape(np.ravel(data),(parameters["PixelArea"][0],parameters["PixelArea"][2]))
    
    def asymmetric_window(npoints, centerindex=None, windowtype="blackmanharris"):
        
        if centerindex is None:
            centerindex = int(len(windowPart2)/2)

        # Calculate the length of the two sides
        length1 = (centerindex)*2
        length2 = (npoints-centerindex)*2

        # Generate the two parts of the window

        windowfunc = getattr(signal.windows,windowtype)
        windowPart1 = windowfunc(length1)
        windowPart2 = windowfunc(length2)

        # Construct the asymetric window from the two sides
        asymWindow1 = windowPart1[0:int(len(windowPart1)/2)]
        if npoints % 2 == 0:
            asymWindow2 = windowPart2[int(len(windowPart2)/2):int(len(windowPart2))]
        else:
            asymWindow2 = windowPart2[int(len(windowPart2)/2+1):int(len(windowPart2))]

        return np.concatenate((asymWindow1, asymWindow2))
    
    def analyse_steps(maxis):
        stepsize = np.zeros((np.shape(maxis)[0],1))
        stepspread = np.zeros((np.shape(maxis)[0],1))
        for i in range(np.shape(maxis)[0]):
            stepsize[i] = np.mean(np.diff(maxis[i,:]))
            stepspread[i] = np.std(np.diff(maxis[i,:]))

        return stepsize, stepspread


"""
Module pySNOM
=============
Scanning Near-field Optical Microscopy (SNOM) analysis tools
"""
from .defaults import defaults
from .images import Image, Transformation
from .readers import Reader
from .spectra import NeaSpectrum
from .interferograms import NeaInterferogram

__version__ = '0.1.0'
__author__ = 'Gergely Németh, Ferenc Borondics'
__credits__ = 'Wigner Research Centre for Physics, Synchrotron SOLEIL'
__all__ = ["defaults","Reader","Image","Transformation","NeaSpectrum","NeaInterferogram"]



import unittest
import os

import numpy as np

import pySNOM
from pySNOM import readers, Image

class TestReaders(unittest.TestCase):
    def test_gwyread(self):
        f = 'datasets/testPsHetData.gwy'
        file_reader = readers.GwyReader(fullfilepath=os.path.join(pySNOM.__path__[0], f),channelname='O3A raw')
        gwy_image = file_reader.read()

        np.testing.assert_almost_equal(gwy_image.data[7][7], 15.213262557983398)
        np.testing.assert_almost_equal(gwy_image.data[8][8], 15.736936569213867)
        np.testing.assert_almost_equal(gwy_image.data[9][9], 13.609171867370605)

    def test_readgsf(self):
        f = 'datasets/testPsHet O3A raw.gsf'
        file_reader = readers.GsfReader(os.path.join(pySNOM.__path__[0], f))
        gwy_image = file_reader.read()

        np.testing.assert_almost_equal(gwy_image.data[7][7], 15.213262557983398)
        np.testing.assert_almost_equal(gwy_image.data[8][8], 15.736936569213867)
        np.testing.assert_almost_equal(gwy_image.data[9][9], 13.609171867370605)

    def test_inforead(self):
        f = 'datasets/testinfofile.txt'
        file_reader = readers.NeaInfoReader(os.path.join(pySNOM.__path__[0], f))
        data = file_reader.read()

        np.testing.assert_approx_equal(data["ScanArea"][0],5.000)
        np.testing.assert_almost_equal(data["TargetWavelength"], 6.006)
        np.testing.assert_string_equal(data["DemodulationMode"], "PsHet")

    def test_spectrumread(self):
        f = 'datasets/testspectrum_singlepoint.txt'
        file_reader = readers.NeaSpectrumReader(os.path.join(pySNOM.__path__[0], f))
        data, params = file_reader.read()

        np.testing.assert_almost_equal(data['O2A'][0], 0.1600194)
        np.testing.assert_string_equal(params["Scan"], "Fourier Scan")

    def test_interferogramread(self):
        f = 'datasets/testifg_singlepoint.txt'
        file_reader = readers.NeaInterferogramReader(os.path.join(pySNOM.__path__[0], f))
        data, params = file_reader.read()

        np.testing.assert_almost_equal(data['O2A'][0], 9.580825)
        np.testing.assert_almost_equal(params["Regulator"][0], 3.767854)
        np.testing.assert_string_equal(params["Scan"], "Fourier Scan")

if __name__ == '__main__':
    unittest.main()



import numpy as np
import copy
from enum import Enum
from pySNOM.defaults import Defaults
from gwyfile.objects import GwyDataField

MeasurementModes = Enum('MeasurementModes', ['None','AFM', 'PsHet', 'WLI', 'PTE', 'TappingAFMIR', 'ContactAFM'])
DataTypes = Enum('DataTypes', ['Amplitude', 'Phase', 'Topography'])
ChannelTypes = Enum('ChannelTypes', ['None','Optical','Mechanical'])

# Full measurement data containing all the measurement channels
class Measurement:
    def __init__(self, data, filename=None, info=None, mode="None"):
        self.filename = filename # Full path with name
        self.mode = mode # Measurement mode (PTE, PSHet, AFM, NanoFTIR) - Enum MeasurementModes
        self._data = data # All channels - Dictionary
        self.info = info

    @property
    def mode(self):
        """Property - measurement mode (Enum)"""
        return self._mode
    @mode.setter
    def mode(self, value: str):
        try:
            self._mode = MeasurementModes[value]
        except ValueError:
            self._mode = MeasurementModes["AFM"]
            raise ValueError(value + 'is not a measurement mode!')

    @property
    def data(self):
        """Property - data (dict with GwyDataFields)"""
        return self._data

    @property
    def info(self):
        """Property - info (dictionary)"""
        return self._info
    
    @info.setter
    def info(self, info):
        self._info = info
        if not info == None:
            m = self._info["Scan"]
            self.mode = Defaults().image_mode_defs[m]

    # METHODS --------------------------------------------------------------------------------------
    def extract_channel(self, channelname: str):
        """Returns a single data channel as GwyDataField"""
        channel = self.data[channelname]
        return channel

    def image_from_channel(self, channelname: str):
        """Returns a single Image object with the requred channeldata"""
        channeldata = self.extract_channel(channelname)
        singleimage = GwyImage(channeldata, filename = self.filename, mode = self.mode, channel = channelname, info = self.info)

        return singleimage


# Single image from a single data channel
class Image(Measurement):
    def __init__(self, data, filename=None, mode="AFM", channelname='Z raw', order=0, datatype=DataTypes['Topography'], info=None):
        super().__init__(data, filename, info=info, mode=mode)
        # Describing channel and datatype
        self.channel = channelname # Full channel name
        self.order = int(order)   # Order, nth
        self.datatype = datatype # Amplitude, Phase, Topography - Enum DataTypes
        
        self.data = data
        self.xoff = 0
        self.yoff = 0
        self.xreal = 1
        self.yreal = 1

    @property
    def data(self):
        """Property - data (numpy array)"""
        # Set the data
        return self._data
    
    @data.setter
    def data(self, new_data):
        """Setter for data (optional if you want data to be modifiable later)"""
        self._data = new_data
        self.xres, self.yres = np.shape(new_data)

    @property
    def channel(self):
        """Property - channel (string)"""
        return self._channel
    
    @channel.setter
    def channel(self,value):

        self._channel = value
        self.channeltype, self.order, self.datatype = type_from_channelname(value)

class GwyImage(Image):
    def __init__(self, data, filename=None, mode="AFM", channelname='Z raw', order=0, datatype=DataTypes['Topography'], info=None):
        super().__init__(data, filename=filename, mode=mode, channelname=channelname, order=order, datatype=datatype, info=info)
    
        self.data = data
        self.xoff = data.xoff
        self.yoff = data.yoff
        self.xreal = data.xreal
        self.yreal = data.yreal

    @property
    def data(self):
        """Property - data (numpy array)"""
        # Set the data
        return self._data
    
    @data.setter
    def data(self,value):
        self._data = value.data
        self.xres, self.yres = np.shape(self._data)
        if self._data is None:
            raise ValueError("The provided data object does not contain 'data' attribute")

def type_from_channelname(channelname):
    if channelname[0] == 'O':
        channeltype = ChannelTypes["Optical"]
    elif 'M' in channelname:
        channeltype = ChannelTypes["Mechanical"]
    else:
        channeltype = ChannelTypes["None"]

    if 'Z' in channelname:
        order = 0
    else:
        order = int(channelname[1])
    
    if channelname[2] == 'A':
        datatype = DataTypes["Amplitude"]
    elif 'Z' in channelname:
        datatype = DataTypes["Topography"]
    else:
        datatype = DataTypes["Phase"]

    return channeltype, order, datatype

class Transformation:

    def transform(self, data):
        raise NotImplementedError()


class LineLevel(Transformation):

    def __init__(self, method='median', datatype=DataTypes.Phase):
        self.method = method
        self.datatype = datatype

    def transform(self, data):
        if self.method == 'median':
            norm = np.nanmedian(data, axis=1, keepdims=True)
        elif self.method == 'mean':
            norm = np.nanmean(data, axis=1, keepdims=True)
        elif self.method == 'difference':
            if self.datatype == DataTypes.Amplitude:
                norm = np.nanmedian(data[1:] / data[:-1], axis=1, keepdims=True)
            else:
                norm = np.nanmedian(data[1:] - data[:-1], axis=1, keepdims=True)
            data = data[:-1]  # difference does not make sense for the last row
        else:
            if self.datatype == DataTypes.Amplitude:
                norm = 1
            else:
                norm = 0

        if self.datatype == DataTypes.Amplitude:
            return data / norm
        else:
            return data - norm

class RotatePhase(Transformation):

    def __init__(self, degree=0.0):
        self.degree = degree

    def transform(self, data):
        # Construct complex dataset
        complexdata = np.exp(data*complex(1j))
        # Rotate and extract phase
        return np.angle(complexdata*np.exp(np.deg2rad(self.degree)*complex(1j)))

class SelfReference(Transformation):

    def __init__(self, referencedata=1, datatype=DataTypes.Phase):
        self.datatype = datatype
        self.referencedata = referencedata
    def transform(self, data):
        if self.datatype == DataTypes.Amplitude:
            return data / self.referencedata
        elif self.datatype == DataTypes.Phase:
            return data - self.referencedata
        else:
            raise RuntimeError("Self-referencing makes only sense for amplitude or phase data")

class SimpleNormalize(Transformation):

    def __init__(self, method='median', value=1.0, datatype=DataTypes.Phase):
        self.method = method
        self.value = value
        self.datatype = datatype

    def transform(self, data):
        match self.method:
            case 'median':
                if self.datatype == DataTypes.Amplitude:
                    return data / np.nanmedian(data)
                else:
                    return data - np.nanmedian(data)
            case 'mean':
                if self.datatype == DataTypes.Amplitude:
                    return data / np.nanmean(data)
                else:
                    return data - np.nanmean(data)
            case 'manual':
                if self.datatype == DataTypes.Amplitude:
                    return data / self.value
                else:
                    return data - self.value
                
class BackgroundPolyFit(Transformation):

    def __init__(self, xorder=1, yorder=1, datatype=DataTypes.Phase):
        self.xorder = int(xorder)
        self.yorder = int(yorder)
        self.datatype = datatype
        
    def transform(self, data):
        Z = copy.deepcopy(data)
        x = list(range(0, Z.shape[1]))
        y = list(range(0, Z.shape[0]))
        X, Y = np.meshgrid(x, y)
        x, y = X.ravel(), Y.ravel()
        b = Z.ravel()
        notnanidxs = np.argwhere(~np.isnan(b))
        b = np.ravel(b[notnanidxs])
        x = np.ravel(x[notnanidxs])
        y = np.ravel(y[notnanidxs])

        def get_basis(x, y, max_order_x=1, max_order_y=1):
            """Return the fit basis polynomials: 1, x, x^2, ..., xy, x^2y, ... etc."""
            basis = []
            for i in range(max_order_y+1):
                # for j in range(max_order_x - i +1):
                for j in range(max_order_x+1):
                    basis.append(x**j * y**i)
            return basis

        try:
            basis = get_basis(x, y, self.xorder, self.yorder)
            A = np.vstack(basis).T
            c, r, rank, s = np.linalg.lstsq(A, b, rcond=None)

            background = np.sum(c[:, None, None] * np.array(get_basis(X, Y, self.xorder, self.yorder)).reshape(len(basis), *X.shape),axis=0)
        except ValueError:
                background = np.ones(np.shape(data))
                print("X and Y order must be integer!")

        if self.datatype == DataTypes["Amplitude"]:
            return Z / background, background
        else:
            return Z-background, background